# Copyright 2021-2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

if { [info exists ::env(VDD_NETS)] || [info exists ::env(GND_NETS)] } {
    # they both must exist and be equal in length
    # current assumption: they cannot have a common ground
    if { ! [info exists ::env(VDD_NETS)] || ! [info exists ::env(GND_NETS)] } {
        puts stderr "\[ERROR\] VDD_NETS and GND_NETS must *both* either be defined or undefined"
        exit -1
    }
    set ::env(VDD_NET) [lindex $::env(VDD_NETS) 0]
    set ::env(GND_NET) [lindex $::env(GND_NETS) 0]

} else {
    set ::env(VDD_NET) $::env(VDD_PIN)
    set ::env(GND_NET) $::env(GND_PIN)
    set ::env(VDD_NETS) $::env(VDD_PIN)
    set ::env(GND_NETS) $::env(GND_PIN)
}
